#ifndef FrameCPP__COMMON__IOStream_HH
#define FrameCPP__COMMON__IOStream_HH

#if !defined(SWIGIMPORTED)
#include <iomanip>
#include <iostream>
#include <fstream>
#include <list>
#include <memory>
#include <string>
#include <typeinfo>
#include <vector>
#endif /* !defined(SWIGIMPORTED) */

#include "general/autoarray.hh"
#include "general/fstream.hh"
#include "general/types.hh"
#include "general/reverse.hh"
#include "general/gpstime.hh"
#include "general/util.hh"
#if !defined(SWIGIMPORTED)
#include "general/Profile.hh"
#include "general/SharedPtr.hh"
#include "general/System.hh"
#endif /* !defined(SWIGIMPORTED) */

#include "framecpp/Common/FrameSpec.hh"

#include "framecpp/Common/CheckSum.hh"
#include "framecpp/Common/CheckSumFilter.hh"
#include "framecpp/Common/Dictionary.hh"
#include "framecpp/Common/MD5SumFilter.hh"
#include "framecpp/Common/PTR_STRUCT_BASE.hh"
#include "framecpp/Common/StreamFilter.hh"

#define FRAME_STREAM_FILTER_IN_BUFFER 0

#define LM_DEBUG_OUTPUT 0
#define LM_DEBUG_INPUT 	0

#if LM_DEBUG_OUTPUT || LM_DEBUG_INPUT
static bool filter_on = true;/* false */
#endif /* LM_DEBUG_OUTPUT || LM_DEBUG_INPUT */

namespace FrameCPP
{
  namespace Common
  {
    class Dictionary;
    class StreamRefInterface;
    class PTR_STRUCT_BASE;
    class FrStatData;

    template< typename T > class Container;
    template < typename L > class STRING;
    template < typename T, const std::string& (T::*F)() const > class SearchContainer;

    //===================================================================
    // Forward declaration of some templated classes
    //===================================================================
    class IStream;
    class OStream;
    
    //===================================================================
    // \brief Interface definition to Frame Buffer
    //===================================================================
    class FrameBufferInterface
    {
    public:
      typedef std::streambuf::char_type	char_type;
      typedef INT_4U buffer_size_type;

      /// \brief The default size for buffered input and output.
      static const buffer_size_type	M_BUFFER_SIZE_DEFAULT;
      static const buffer_size_type	M_BUFFER_SIZE_SYSTEM = ~buffer_size_type( 0 );

      virtual ~FrameBufferInterface( );
      virtual void FilterAdd( FrameCPP::Common::StreamFilter* Filter ) = 0;
      virtual void FilterRemove( FrameCPP::Common::StreamFilter* Filter ) = 0;
      virtual bool FilterInternally( ) const = 0;
      virtual void buffer( ) = 0;
      template< class DT >
      void filter( const DT& Data );
      virtual void filter( const char_type* Start,
			   const char_type* End ) = 0;

      //-----------------------------------------------------------------
      /// \brief Retrieve the string identification of the buffer.
      ///
      /// \return
      ///     An identifier associated with the open buffer.
      //-----------------------------------------------------------------
      const std::string& BufferId( ) const;

    protected:
      void bufferId( const std::string& Id );

    private:
      std::string	m_stream_id;
    };

    inline const std::string& FrameBufferInterface::
    BufferId( ) const
    {
      return m_stream_id;
    }

    inline void FrameBufferInterface::
    bufferId( const std::string& Id )
    {
      m_stream_id = Id;
    }

    template< class DT >
    inline void FrameBufferInterface::
    filter( const DT& Data )
    {
      filter( reinterpret_cast< const char_type* >( &Data ),
	      reinterpret_cast< const char_type* >( &Data + sizeof( Data ) ) );
    }

    //===================================================================
    /// \brief Buffer appropriate for caching frame data
    //===================================================================
    template< typename BT >
    class FrameBuffer
      : public FrameBufferInterface,
	public BT
    {
    public:
      /// \brief Type of self
      typedef FrameBuffer< BT > buffer_type;

      /// \brief type representing a character in the stream.
      typedef typename BT::char_type char_type;

      /// \brief integer type
      typedef typename BT::int_type int_type;
      /// \brief type appropriate for holding stream position information
      typedef typename BT::pos_type pos_type;
      /// \brief type apprpriate to hold stream offset request
      typedef typename BT::off_type off_type;

      using BT::pubsetbuf;

      using BT::pbase;
      using BT::pptr;
      using BT::epptr;

      using BT::eback;
      using BT::gptr;
      using BT::egptr;

      /// \brief constructor
      FrameBuffer( std::ios::openmode Mode );

      /// \brief adds a filter to the stream
      void FilterAdd( FrameCPP::Common::StreamFilter* Filter );
      void FilterRemove( FrameCPP::Common::StreamFilter* Filter );

      /// \brief Returns true if filtering happens internally; false otherwise
      bool FilterInternally( ) const;

      FrameBuffer<BT>* close( );

      FrameBuffer<BT>* open( const std::string& Filename, std::ios::openmode Mode );
      void UseMemoryMappedIO( bool Value );

    protected:
      friend class IStream;
      friend class OStream;

      /// \brief Routine to register if the caller has specified a buffer
      virtual FrameBuffer< BT >* setbuf( char_type* S, std::streamsize N );

      /// \brief Establish buffering
      virtual void buffer( );

    private:
      /// \brief Type specifying the container holding the active filters.
      typedef std::list< FrameCPP::Common::StreamFilter* > filters_type;

      /// \brief Establish the size of the current user buffer.
      void set_user_buf( const char_type* S, std::streamsize N );

      /// \brief Mark the current position in the buffer
      void mark( const char_type* Offset = (const char_type*)0 );

      template< typename DT >
      void filter( const DT& Data );

      /// \brief Perform filtering operations on the stream
      void filter( const typename BT::char_type* Buffer, const INT_8U Length );

      /// \brief Perform filtering operations on the stream
      void filter( const typename BT::char_type* Begin,
		   const typename BT::char_type* End );

      /// \brief State of user supplied buffer
      bool						m_buffer_user_supplied;

      General::AutoArray< typename BT::char_type >	m_buffer;

      /// \brief Container holding all active filters
      filters_type	m_filters;

      /// \brief Size of the internal buffer
      std::streamsize	m_buffer_size;

      /// \brief Location of the user specified buffer
      const char_type*	m_user_buf;
      /// \brief Size of the user specified buffer
      std::streamsize	m_user_buf_size;

      INT_8U m_bytes_filtered;
      
      std::ios::openmode	m_mode;
    };

    template< typename BT >
    inline void FrameBuffer< BT >::
    set_user_buf( const char_type* S, std::streamsize N )
    {
      m_user_buf = S;
      m_user_buf_size = N;
    }

    template< typename BT >
    template< typename DT >
    inline void FrameBuffer< BT >::
    filter( const DT& Data )
    {
      filter( &Data, sizeof( Data ) );
    }

    template< typename BT >
    inline void FrameBuffer< BT >::
    filter( const char_type* Begin,
	    const char_type* End )
    {
      if ( ( Begin == (const char_type*)0 )
	   || ( Begin == End ) )
      {
	return;
      }

      m_bytes_filtered += ( End - Begin ) * sizeof( char_type );

#if LM_DEBUG_OUTPUT || LM_DEBUG_INPUT
      if ( filter_on )
      {
	MD5Sum m;
	m.Update( (const MD5Sum::data_type)Begin,
		  MD5Sum::size_type( End - Begin ) );
	m.Finalize( );

	std::cerr << "FrameBuffer< BT >::filter:"
		  << " Begin: " << (void*)Begin
		  << " End: " << (void*)End
		  << " m_bytes_filterd: " << m_bytes_filtered
		  << " md5sum: " << m
		  << std::endl;
      }
#endif /* LM_DEBUG_OUTPUT || LM_DEBUG_INPUT */

      for( filters_type::const_iterator
	     cur = m_filters.begin( ),
	     last = m_filters.end( );
	   cur != last;
	   ++cur )
      {
	if ( *cur )
	{
	  INT_8U length = End - Begin;
	  const char_type*	start = Begin;

	  if ( FilterInternally( ) )
	  {
	    const char_type*	pos
	      = reinterpret_cast< const char_type*>( (*cur)->Offset( ) );

	    if ( m_mode & std::ios::in )
	    {
	      if ( ( start >= eback( ) )
		   && ( start <= egptr( ) )
		   && ( pos >= start )
		   && ( pos <= egptr( ) ) )
	      {
		start = ( pos ) ? pos : eback( );
		length = End - start;
	      }
	    }
	    else if ( m_mode & std::ios::out )
	    {
	      if ( ( start >= pbase( ) )
		   && ( start <= epptr( ) )
		   && ( pos >= start )
		   && ( pos <= epptr( ) ) )
	      {
		start = ( pos ) ? pos : pbase( );
		length = End - start;
	      }
	    }
	  } // if ( FilterInternally( ) )
#if LM_DEBUG_OUTPUT || LM_DEBUG_INPUT
	  if ( filter_on )
	  {
	    std::cerr << "DEBUG: filter:"
		      << " m_bytes_filterd: " << m_bytes_filtered
		      << " start: " << (void*)start
		      << " length: " << length
		      << std::endl;
	  }
#endif /* LM_DEBUG_OUTPUT || LM_DEBUG_INPUT */
	  (*cur)->pubfilter( static_cast< const char* >( start ),
			     length * sizeof( char_type ) );
	}
	mark( );
      }
    }

    template< typename BT >
    inline void FrameBuffer< BT >::
    filter( const typename BT::char_type* Buffer, const INT_8U Length )
    {
      return filter( Buffer, Buffer + Length );
    }

    template< typename BT >
    FrameBuffer< BT >::
    FrameBuffer( std::ios::openmode Mode )
      : m_buffer_user_supplied( false ),
	m_buffer_size( 0 ),
	m_bytes_filtered( 0 ),
	m_mode( Mode )
    {
    }

    template< typename BT >
    void FrameBuffer< BT >::
    FilterAdd( FrameCPP::Common::StreamFilter* Filter )
    {
      if ( Filter )
      {
	if ( FilterInternally( ) )
	{
	  if ( m_mode & std::ios::in )
	  {
	    Filter->Offset( gptr( ) );
	  }
	  else if ( m_mode & std::ios::out )
	  {
	    Filter->Offset( pptr( ) );
	  }
	} // if ( FilterInternally( ) )
	m_filters.push_back( Filter );
      }
    }

    template< typename BT >
    void FrameBuffer< BT >::
    FilterRemove( FrameCPP::Common::StreamFilter* Filter )
    {
      filters_type::size_type s = m_filters.size( );
      m_filters.remove( Filter );
      if ( FilterInternally( ) )
      {
	if ( m_filters.size( ) < s )
	{
	  const char_type* offset
	    = reinterpret_cast< const char_type* >( Filter->Offset( ) );
	  
	  if ( m_mode & std::ios::in )
	  {
	    if ( offset == ( const char_type* )0 )
	    {
	      offset = eback( );
	    }
	    Filter->pubfilter( offset, gptr( ) - offset );
	  }
	  else if ( m_mode & std::ios::out )
	  {
	    if ( offset == ( const char_type* )0 )
	    {
	      offset = pbase( );
	    }
	    Filter->pubfilter( offset, pptr( ) - offset );
	  }
	}
      } // if ( FilterInernally( )
    }

    template< typename BT >
    FrameBuffer< BT >* FrameBuffer< BT >::
    setbuf( char_type* S, std::streamsize N )
    {
      if ( N != std::streamsize( M_BUFFER_SIZE_SYSTEM ) )
      {
	m_buffer_user_supplied = true;
	BT::setbuf( S, N );
      }
      return this;
    }

    template< typename BT >
    void FrameBuffer< BT >::
    buffer( )
    {
      if ( m_buffer_user_supplied == false )
      {
	//---------------------------------------------------------------
	// Allocation of the default buffer
	//---------------------------------------------------------------
	m_buffer.reset( new typename BT::char_type[ M_BUFFER_SIZE_DEFAULT ] );
	BT::setbuf( m_buffer.get( ), M_BUFFER_SIZE_DEFAULT );
      }
    }

    //===================================================================
    //===================================================================
    class StreamBase
    {
    public:
      typedef General::SharedPtr< FrameSpec::Object >	object_type;
      typedef const Dictionary::streamref_type::element_type*
      const_stream_ref_ptr_type;
      typedef Dictionary::streamref_type	streamref_type;

      StreamBase( );

      virtual ~StreamBase( );

      virtual void Cleanup( );

      INT_2U GetClassId( const std::string& ClassName ) const;

      INT_2U GetFSIId( INT_2U StreamId ) const;

      cmn_streamsize_type PtrStructBytes( ) const;

      const_stream_ref_ptr_type
      ReferenceStreamRef( ) const;

      const Dictionary& GetDictionary( ) const;

      Dictionary& GetDictionary( );

    protected:
      typedef General::unordered_map< INT_2U, INT_2U >
      streamid_to_fsiid_type;

      typedef General::SharedPtr< PTR_STRUCT_BASE >	ptr_struct_base_type;

      Dictionary				m_dictionary;
      object_type				m_null_stream_ref;
      ptr_struct_base_type			m_ptr_struct_object;
      mutable streamid_to_fsiid_type		m_stream_id_to_fsi_id;

      //-----------------------------------------------------------------
      /// \brief Checksum for file
      //-----------------------------------------------------------------
      std::auto_ptr< CheckSumFilter >		m_file_checksum;
      //-----------------------------------------------------------------
      /// \brief Checksum for frame
      //-----------------------------------------------------------------
      std::auto_ptr< CheckSumFilter >		m_frame_checksum;
      //-----------------------------------------------------------------
      /// \brief Checksum for object
      //-----------------------------------------------------------------
      std::auto_ptr< CheckSumFilter >		m_obj_checksum;
      //-----------------------------------------------------------------
      /// \brief MD5Sum value for the files
      //-----------------------------------------------------------------
      std::auto_ptr< MD5SumFilter >		m_file_md5sum;

      
      const FrameSpec::Info&	frameSpecInfo( ) const;

      virtual void frameSpecInfo( const FrameSpec::Info* Spec );

    private:
      typedef General::unordered_map< std::string, INT_2U,
				      General::hash< std::string >,
				      CaseInsensitiveCmp >
      cntocid_mapping_type;

      const FrameSpec::Info*		m_frame_spec;
      mutable cmn_streamsize_type	m_ptr_struct_bytes;
      CheckSum*				m_crc_file;
      CheckSum*				m_crc_frame;
      CheckSum*				m_crc_obj;

      static const bool			m_initialized;
      static cntocid_mapping_type	m_class_name_to_class_id;

      static bool initialize( );
    };

    inline cmn_streamsize_type StreamBase::
    PtrStructBytes( ) const
    {
      if ( m_ptr_struct_bytes == 0 )
      {
	const FrameSpec::Object* o
	  = frameSpecInfo( ).FrameObject( FrameSpec::Info::FSI_PTR_STRUCT );
	if ( o )
	{
	  m_ptr_struct_bytes = o->Bytes( *this );
	}
      }
      return m_ptr_struct_bytes;
    }

    inline const Dictionary& StreamBase::
    GetDictionary( ) const
    {
      return m_dictionary;
    }

    inline Dictionary& StreamBase::
    GetDictionary( )
    {
      return m_dictionary;
    }

    inline INT_2U StreamBase::
    GetFSIId( INT_2U StreamId ) const
    {
      const streamid_to_fsiid_type::const_iterator
	sid_iter = m_stream_id_to_fsi_id.find( StreamId );
      if ( sid_iter == m_stream_id_to_fsi_id.end( ) )
      {
	//---------------------------------------------------------------
	// Report the error
	//---------------------------------------------------------------
	std::ostringstream	msg;
	msg << "Unable to map from the stream class id (" << StreamId
	    << ") to internal class id"
	  ;
	throw
	  std::range_error( msg.str( ) );
      }
      return sid_iter->second;
    }

    inline const FrameSpec::Info& StreamBase::
    frameSpecInfo( ) const
    {
      return *m_frame_spec;
    }

    //===================================================================
    //===================================================================
    class IStream
      : virtual public StreamBase,
	public std::istream
    {
    public:
      typedef std::istream::char_type char_type;
      typedef std::istream istream_type;
      typedef FrameBufferInterface buffer_type;

      IStream( buffer_type* Buffer );

      virtual ~IStream( );

      virtual void Cleanup( );

      CheckSumFilter* GetCheckSumFile( ) const;

      void SetCheckSumFile( const CheckSum::kind_type Type );

      CheckSumFilter* GetCheckSumFrame( ) const;

      void SetCheckSumFrame( const CheckSum::kind_type Type );

      CheckSumFilter* GetCheckSumObject( ) const;

      void SetCheckSumObject( const CheckSum::kind_type Type );

      Dictionary::ptr_struct_key_type
      NextLookup( object_type Obj ) const;

      void RemoveResolver( const void * Old, int MaxDepth );

      template< typename T >
      void PushPtr( General::SharedPtr< T >* Obj );

      template< typename T_NEW, typename T_OLD >
      void ReplacePtr( General::SharedPtr< T_NEW >* Obj,
		       General::SharedPtr< T_OLD >* OldPtr,
		       const int MaxDepth );

      template< typename T > void PushRef( Container< T >& C );
      template< typename T > void ReplaceRef( Container< T >& C,
					      const void * Old,
					      const int MaxDepth );

      FrameBufferInterface* rdbuf( ) const;

      IStream& read( char_type* S, std::streamsize N );

      virtual ptr_struct_base_type Next( object_type Obj );

      virtual ptr_struct_base_type Next( object_type::element_type* Obj );

      ptr_struct_base_type ReadPtrStruct( );

      void Resolve( );

      template< typename T > IStream&
      operator>>( General::SharedPtr< T >* Data );

      template< typename T, const std::string& (T::*F)() const > IStream&
      operator>>( SearchContainer< T, F >& SC );

      template< typename T > IStream&
      operator>>( Container< T >& C );

      template< typename L > IStream&
      operator>>( STRING< L >& String );

      template< typename L > IStream&
      operator>>( std::vector< STRING< L > >& String );

      template< typename DT >
      IStream& operator>>( std::vector< DT >& Data );

      template< typename DT >
      IStream& operator>>( DT& Data );

      bool ByteSwapping( ) const;

    protected:
      class resolver
      {
      public:
	resolver( ptr_struct_base_type PtrStruct );
	resolver( resolver& Source );
	virtual ~resolver( );

	virtual void Resolve( IStream& Stream ) = 0;
	virtual bool operator==( const void* RHS ) const = 0;

	ptr_struct_base_type PtrStruct( ) const;

      protected:
	ptr_struct_base_type	m_ptr_struct;
      };

      typedef General::SharedPtr< resolver > resolver_type;
      typedef std::list< resolver_type > resolver_container_type;

      bool	m_logNextPtr;

      void nextReplace( object_type OldObj, object_type NewObj );

      void byteSwapping( bool Value );


      virtual void pushResolver( resolver_type Resolver );

      object_type objectLookup( object_type::element_type* Pointer );

      const resolver_container_type& resolverContainer( ) const;

      resolver_container_type& resolverContainer( );

      void setAutoDeleteBuffer( const bool Value );

    private:
      template< typename T >
      class ptr_reference
	: public resolver
      {
      public:
	ptr_reference( General::SharedPtr< T >* Obj, ptr_struct_base_type PtrStruct );
	ptr_reference( General::SharedPtr< T >* Obj, resolver& Source );
	virtual ~ptr_reference( );

	virtual void Resolve( IStream& Stream );

	virtual bool operator==( const void* RHS ) const;

      private:
	General::SharedPtr< T >*	m_object;
      };

      template< typename T >
      class container_reference
	: public resolver
      {
      public:
	container_reference( Container< T >& C, ptr_struct_base_type PtrStruct );
	container_reference( Container< T >& C, resolver& Source );
	~container_reference( );

	virtual void Resolve( IStream& Stream );

	virtual bool operator==( const void* RHS ) const;

      private:
	Container< T >&		m_container;
      };

      template < typename T >
      struct smart_pointer_hash
      {
	inline size_t operator( )(const T Key ) const
	{
	  return reinterpret_cast< size_t >( Key.get( ) );
	}
      };

      /// \brief Maintains information to reconstruct linked lists
      typedef General::unordered_map< object_type,
				      Dictionary::ptr_struct_key_type,
				      smart_pointer_hash< object_type > >
      next_container_type;

      typedef General::unordered_map< object_type::element_type*,
				      object_type,
				      General::hash_pointer< object_type::element_type > >
      object_container_type;

      std::auto_ptr< buffer_type >	m_buffer;
      bool				m_buffer_auto_delete;
      bool				m_filter_internally;
      next_container_type		m_next_container;
      bool				m_byte_swapping;
      resolver_container_type		m_resolver_container;
      std::auto_ptr< CheckSumFilter >	m_file_checksum;
      std::auto_ptr< CheckSumFilter >	m_frame_checksum;
      std::auto_ptr< CheckSumFilter >	m_obj_checksum;
      object_container_type		m_objects;
    };

    inline bool IStream::
    ByteSwapping( ) const
    {
      return m_byte_swapping;
    }

    inline Dictionary::ptr_struct_key_type IStream::
    NextLookup( object_type Obj ) const
    {
      next_container_type::const_iterator
	key_iterator = m_next_container.find( Obj );
	
      if ( key_iterator != m_next_container.end( ) )
      {
	return key_iterator->second;
      }
      Dictionary::ptr_struct_key_type key( 0, 0 );
      return key;
    }

    inline void IStream::
    nextReplace( object_type Old, object_type New )
    {
      next_container_type::const_iterator
	key_iterator = m_next_container.find( Old );
	
      if ( key_iterator != m_next_container.end( ) )
      {
	m_next_container[ New ] = key_iterator->second;
	m_next_container.erase( Old );
      }
    }

    inline void IStream::
    byteSwapping( bool Value )
    {
      m_byte_swapping = Value;
    }

    inline IStream::object_type IStream::
    objectLookup( object_type::element_type* Pointer )
    {
      object_type	retval;

      object_container_type::const_iterator
	pos = m_objects.find( Pointer );
      if ( pos == m_objects.end( ) )
      {
	retval.reset( Pointer );

	m_objects[ Pointer ] = retval;
      }
      else
      {
	retval = pos->second;
      }
      return retval;
    }

    inline const IStream::resolver_container_type& IStream::
    resolverContainer( ) const
    {
      return m_resolver_container;
    }

    inline IStream::resolver_container_type& IStream::
    resolverContainer( )
    {
      return m_resolver_container;
    }

    inline void IStream::
    setAutoDeleteBuffer( const bool Value )
    {
      m_buffer_auto_delete = Value;
    }

    //-------------------------------------------------------------------
    //-------------------------------------------------------------------
    inline IStream::resolver::
    resolver( resolver& Source )
      : m_ptr_struct( Source.m_ptr_struct )
    {
    }

    inline IStream::ptr_struct_base_type IStream::resolver::
    PtrStruct( ) const
    {
      return m_ptr_struct;
    }

    //===================================================================
    // class IStream::ptr_reference
    //===================================================================

    template< typename T >
    inline IStream::ptr_reference< T >::
    ptr_reference( General::SharedPtr< T >* Obj, resolver& Source )
      : resolver( Source ),
	m_object( Obj )
    {
    }

    
    template< typename T >
    IStream::ptr_reference< T >::
    ptr_reference( General::SharedPtr< T >* Ptr, ptr_struct_base_type PtrStruct )
      : resolver( PtrStruct ),
	m_object( Ptr )
    {
      assert( m_object );
    }

    template< typename T >
    IStream::ptr_reference< T >::
    ~ptr_reference( )
    {
      // Nothing to do here
    }

    template< typename T >
    void IStream::ptr_reference< T >::
    Resolve( IStream& Stream )
    {
      //-----------------------------------------------------------------
      // Get all of the objects that are needed to rebuild list.
      //-----------------------------------------------------------------
      Dictionary& dict( Stream.GetDictionary( ) );

      Dictionary::ptr_struct_key_type key( m_ptr_struct->Class( ),
					   m_ptr_struct->Instance( ) );
      Dictionary::streamref_type sr = dict.Ref( key );
      //-----------------------------------------------------------------
      // 
      //-----------------------------------------------------------------
      if ( sr )
      {
	(*m_object)
	  = General::DynamicPointerCast< T >( dict.Ref( sr ) );
      }
    }

    template< typename T >
    bool IStream::ptr_reference< T >::
    operator==( const void* RHS ) const
    {
      return ( m_object == RHS );
    }

    //-------------------------------------------------------------------
    //-------------------------------------------------------------------

    template< typename T >
    IStream::container_reference< T >::
    container_reference( Container< T >& C, ptr_struct_base_type PtrStruct )
      : resolver( PtrStruct ),
	m_container( C )
    {
    }

    template< typename T >
    IStream::container_reference< T >::
    container_reference( Container< T >& C, resolver& Source )
      : resolver( Source ),
	m_container( C )
    {
    }

    template< typename T >
    IStream::container_reference< T >::
    ~container_reference( )
    {
    }

    template< typename T >
    void IStream::container_reference< T >::
    Resolve( IStream& Stream )
    {
      //-----------------------------------------------------------------
      // Get all of the objects that are needed to rebuild list.
      //-----------------------------------------------------------------
      Dictionary& dict( Stream.GetDictionary( ) );

      Dictionary::ptr_struct_key_type key( m_ptr_struct->Class( ),
					   m_ptr_struct->Instance( ) );
      Dictionary::streamref_type sr = dict.Ref( key );
      //-----------------------------------------------------------------
      // 
      //-----------------------------------------------------------------
      if ( sr )
      {
	object_type	obj( dict.Ref( sr ) );

	while( obj )
	{
	  //-------------------------------------------------------------
	  // Place object into container
	  //-------------------------------------------------------------
	  dict.Remove( obj );
	  General::SharedPtr< T >
	    c_obj = General::DynamicPointerCast< T >( obj );
	  if ( c_obj )
	  {
	    m_container.append( c_obj );
	  }
	  //-------------------------------------------------------------
	  // Advance to the next object
	  //-------------------------------------------------------------
	  key = Stream.NextLookup( obj );
	  if ( ( key.first == 0 ) && ( key.second == 0 ) )
	  {
	    break;
	  }
	  sr = dict.Ref( key );
	  if ( sr )
	  {
#if defined( SharedPtr )
	    obj = dict.Ref( sr );
#else
	    obj
	      = dynamic_cast< T* >
	      ( const_cast< Dictionary::object_type* >( dict.Ref( sr ) ) );
#endif
	  }
	  else
	  {
#if defined( SharedPtr )
	    obj.reset( );
#else
	    obj = (T*)NULL;
#endif
	  }
	}
      }

    }

    template< typename T >
    bool IStream::container_reference< T >::
    operator==( const void* RHS ) const
    {
      return ( &m_container == RHS );
    }

    //-------------------------------------------------------------------
    //-------------------------------------------------------------------
    inline IStream::
    IStream( buffer_type* Buffer )
      : std::istream( dynamic_cast< std::streambuf* >( Buffer ) ),
	m_logNextPtr( true ),
	m_buffer( Buffer ),
	m_buffer_auto_delete( true )
    {
      if ( Buffer )
      {
	m_filter_internally = Buffer->FilterInternally( );
      }
    }

    inline IStream::
    ~IStream( )
    {
      if ( m_buffer_auto_delete == false )
      {
	m_buffer.release( );
      }
    }

    inline CheckSumFilter* IStream::
    GetCheckSumFile( ) const
    {
      return m_file_checksum.get( );
    }

    inline void IStream::
    SetCheckSumFile( const CheckSum::kind_type Type )
    {
      if ( m_file_checksum.get( ) )
      {
	rdbuf( )->FilterRemove( m_file_checksum.get( ) );
      }
      if ( Type == CheckSum::NONE )
      {
	m_file_checksum.reset( (CheckSumFilter*)NULL );
      }
      else
      {
	m_file_checksum.reset( new CheckSumFilter( Type ) );
	rdbuf( )->FilterAdd( m_file_checksum.get( ) );
      }
    }

    inline CheckSumFilter* IStream::
    GetCheckSumFrame( ) const
    {
      return m_frame_checksum.get( );
    }

    inline void IStream::
    SetCheckSumFrame( const CheckSum::kind_type Type )
    {
      if ( m_frame_checksum.get( ) )
      {
	rdbuf( )->FilterRemove( m_frame_checksum.get( ) );
      }
      if ( Type == CheckSum::NONE )
      {
	m_frame_checksum.reset( (CheckSumFilter*)NULL );
      }
      else
      {
	m_frame_checksum.reset( new CheckSumFilter( Type ) );
	rdbuf( )->FilterAdd( m_frame_checksum.get( ) );
      }
    }

    inline CheckSumFilter* IStream::
    GetCheckSumObject( ) const
    {
      return m_obj_checksum.get( );
    }

    inline void IStream::
    SetCheckSumObject( const CheckSum::kind_type Type )
    {
      if ( m_obj_checksum.get( ) )
      {
	rdbuf( )->FilterRemove( m_obj_checksum.get( ) );
      }
      if ( Type == CheckSum::NONE )
      {
	m_obj_checksum.reset( (CheckSumFilter*)NULL );
      }
      else
      {
	m_obj_checksum.reset( new CheckSumFilter( Type ) );
	rdbuf( )->FilterAdd( m_obj_checksum.get( ) );
      }
    }

    inline void IStream::
    RemoveResolver( const void* Old, int MaxDepth )
    {
      for ( resolver_container_type::iterator
	      cur = m_resolver_container.begin( ),
	      last = m_resolver_container.end( );
	    ( --MaxDepth > 0 ) && ( cur != last );
	    ++cur )
      {
	if ( (*cur) && ( *(*cur) == Old ) )
	{
	  m_resolver_container.erase( cur );
	  break;
	}
      }
    }

    template< typename T >
    inline void IStream::
    PushPtr( General::SharedPtr< T >* Obj )
    {
      //-----------------------------------------------------------------
      // Read the PTR_STRUCT from the stream
      //-----------------------------------------------------------------
      ptr_struct_base_type	ptr_struct( ReadPtrStruct( ) );
      
      if ( ptr_struct->IsNULL( ) )
      {
	return;
      }

      resolver_type	r( new ptr_reference< T >( Obj, ptr_struct ) );

      pushResolver( r );
    }

    template< typename T_NEW, typename T_OLD >
    inline void IStream::
    ReplacePtr( General::SharedPtr< T_NEW >* Obj,
		General::SharedPtr< T_OLD >* OldPtr,
		const int MaxDepth )
    {
      int depth = MaxDepth;

      for ( resolver_container_type::iterator
	      cur = m_resolver_container.begin( ),
	      last = m_resolver_container.end( );
	    ( --depth > 0 ) && ( cur != last );
	    ++cur )
      {
	if ( (*cur) && ( *(*cur) == OldPtr ) )
	{
	  // Create a new reference
	  General::SharedPtr< ptr_reference< T_NEW > >
	    np( new ptr_reference< T_NEW >( Obj, **cur )  );
	  
	  // Out with the old
	  cur = m_resolver_container.erase( cur );
	  // In with the new
	  m_resolver_container.insert( cur, np );
	  break;
	}
      }
    }

    template< typename T >
    inline void IStream::
    PushRef( Container< T >& C )
    {
      //-----------------------------------------------------------------
      // Read the PTR_STRUCT from the stream
      //-----------------------------------------------------------------
      ptr_struct_base_type	ptr_struct( ReadPtrStruct( ) );
	
      if ( ptr_struct->IsNULL( ) )
      {
	//---------------------------------------------------------------
	// No next object
	//---------------------------------------------------------------
	return;
      }

      resolver_type	r( new container_reference< T >( C, ptr_struct ) );

      pushResolver( r );
    }

    template< typename T >
    inline void IStream::
    ReplaceRef( Container< T >& C, const void* Old,
		const int MaxDepth )
    {
      //-----------------------------------------------------------------
      // Go on beyond MaxDepth since most references have a next pointer.
      //-----------------------------------------------------------------
      int depth = MaxDepth + 1;

      for ( resolver_container_type::iterator
	      cur = m_resolver_container.begin( ),
	      last = m_resolver_container.end( );
	    ( --depth > 0 ) && ( cur != last );
	    ++cur )
      {
	if ( (*cur) && ( *(*cur) == Old ) )
	{
	  // Create a new reference
	  General::SharedPtr< container_reference< T > >
	    ncr( new container_reference< T >( C, **cur )  );
	  
	  // Out with the old
	  cur = m_resolver_container.erase( cur );
	  // In with the new
	  m_resolver_container.insert( cur, ncr );
	  break;
	}
      }
    }

    inline IStream::buffer_type* IStream::
    rdbuf( ) const
    {
      return m_buffer.get( );
    }

    inline IStream& IStream::
    read( char_type* S, std::streamsize N )
    {
      istream_type::read( S, N );
      if ( ! m_filter_internally )
      {
	m_buffer->filter( S, S + this->gcount( ) );
      }
      return *this;
    }
    
    template< typename T >
    inline IStream& IStream::
    operator>>( General::SharedPtr< T >* Data )
    {
      PushPtr( Data );

      return *this;
    }

    template< typename T, const std::string& (T::*F)() const >
    inline IStream& IStream::
    operator>>( SearchContainer< T, F >& SC )
    {
      return SC.StreamIn( *this );
    }
    
    template< typename T >
    inline IStream& IStream::
    operator>>( Container< T >& C )
    {
      return C.StreamIn( *this );
    }

    template< typename L >
    inline IStream& IStream::
    operator>>( STRING< L >& String )
    {
      String.Read( *this );
      return *this;
    }

    template< typename L >
    inline IStream& IStream::
    operator>>( std::vector< STRING< L > >& Strings )
    {
      for( typename std::vector< STRING< L > >::iterator
	     cur = Strings.begin( ),
	     last = Strings.end( );
	   cur != last;
	   ++cur )
      {
	cur->Read( *this );
      }
      return *this;
    }

    template< >
    inline IStream& IStream::
    operator>>( General::GPSTime& Data )
    {
      INT_4U	seconds;
      INT_4U	nanoseconds;

      *this >> seconds
	    >> nanoseconds;

      Data = General::GPSTime( seconds, nanoseconds );
      return *this;
    }

    template< typename DT >
    inline IStream& IStream::
    operator>>( std::vector< DT >& Data )
    {
      if ( Data.size( ) > 0 )
      {
	read( reinterpret_cast< IStream::char_type* >( &(Data[0]) ),
	      std::streamsize( sizeof( DT ) * Data.size( ) ) );
	if ( m_byte_swapping )
	{
	  reverse< sizeof( DT ) >( &(Data[ 0 ]), Data.size( ) );
	}
      }
      return *this;
    }

    template< typename DT >
    inline IStream& IStream::
    operator>>( DT& Data )
    {
      read( reinterpret_cast< IStream::char_type* >( &Data ),
	    std::streamsize( sizeof( Data ) ) );
      if (  m_byte_swapping )
      {
	reverse< sizeof( DT ) >( &Data, 1 );
      }
      return *this;
    }

    //===================================================================
    //===================================================================
    class OStream
      : virtual public StreamBase,
	public std::ostream
    {
    public:
      typedef std::ostream::char_type char_type;
      typedef std::ostream ostream_type;
      typedef FrameBufferInterface buffer_type;
      typedef StreamBase::object_type object_type;

      class Pointer
      {
      public:
	Pointer( object_type Data );

	object_type GetPointer( ) const;

      private:
	object_type	m_pointer;
      };

      OStream( buffer_type* Buffer );

      virtual ~OStream( );

      /// param[in,out] CompressionScheme
      /// param[in,out] CompressionType
      void Compression( INT_2U& CompressionScheme, INT_2U& CompressionLevel );

      INT_2U CompressionScheme( ) const;

      INT_2U CompressionLevel( ) const;

      template< typename Header >
      const Header* GetFrHeader( ) const;

      virtual INT_8U GetTOCOffset( ) const;

      void Next( object_type Obj,
		 object_type NextObj );

      object_type Next( object_type Obj ) const;

      const PTR_STRUCT_BASE& Reference( object_type Obj ) const;

      OStream& write( const char_type* S, std::streamsize N );

      /// \brief Get an object from the output queue
      object_type Pop( );

      /// \brief Append an object to the output queue
      void PushSingle( object_type Obj );

      /// \brief Append an object to the output queue
      template< typename Iter >
      void PushMulti( Iter Start,
		      Iter End,
		      size_t Size )
      {
	if ( Size == 0 )
	{
	  return;
	}

	object_queue_type	c;

	object_type  previous;

	if ( Size == 1 )
	{
	  m_dictionary.RefCreate( *Start, *this );
	  m_write_objects_pos
	    = m_write_objects.insert( m_write_objects_pos, *Start );
	  ++m_write_objects_pos;
	}
	else
	{
	  while( Start != End )
	  {
	    m_dictionary.RefCreate( *Start, *this );
	    c.push_back( *Start );
	    if ( previous )
	    {
	      Next( previous, *Start );
	    }
	    previous = *Start;
	    ++Start;
	  }
	  m_write_objects.splice( m_write_objects_pos, c );
	}
      }

      /// \brief Create a PTR_STRUCT for next object
      const Common::PTR_STRUCT_BASE&
      NextPtrStruct( object_type Cur ) const;

      /// \brief Create a PTR_STRUCT for next object
      const Common::PTR_STRUCT_BASE&
      NextPtrStruct( const object_type::element_type* Cur ) const;

      virtual INT_4U GetNumberOfFrames( ) const;

      /// \todo support MD5Sum

      OStream& operator<<( object_type Data );

      OStream& operator<<( Pointer Data );

      template< typename T, const std::string& (T::*F)() const > inline OStream&
      operator<<( const SearchContainer< T, F >& SC )
      {
	return SC.StreamOut( *this );
      }

      template< typename T > inline OStream&
      operator<<( const Container< T >& C )
      {
	return C.StreamOut( *this );
      }

      template< typename L > inline OStream&
      operator<<( const STRING< L >& String )
      {
	String.Write( *this );
	return *this;
      }

      template< typename L > inline OStream&
      operator<<( const std::vector< STRING< L > >& Strings )
      {
	for ( typename std::vector< STRING< L > >::const_iterator
		cur = Strings.begin( ),
		last = Strings.end( );
	      cur != last;
	      ++cur )
	{
	  cur->Write( *this );
	}
	return *this;
      }

      template< typename DT >
      OStream& operator<<( const std::vector< DT >& N );

      template< typename DT >
      OStream& operator<<( const DT& N );

      inline buffer_type* rdbuf( ) const
      {
	return m_buffer.get( );
      }

      CheckSumFilter* GetCheckSumFile( ) const;

      void SetCheckSumFile( const CheckSum::kind_type Type );

      CheckSumFilter* GetCheckSumFrame( ) const;

      void SetCheckSumFrame( const CheckSum::kind_type Type );

      CheckSumFilter* GetCheckSumObject( ) const;

      void SetCheckSumObject( const CheckSum::kind_type Type );

    protected:
      void setFrHeader( std::auto_ptr< FrameSpec::Object >& Obj );

    private:
      typedef std::list< object_type > object_queue_type;

      std::auto_ptr< FrameSpec::Object >	m_file_header;
      std::auto_ptr< buffer_type >		m_buffer;
      bool					m_filter_internally;
      object_queue_type				m_write_objects;
      object_queue_type::iterator		m_write_objects_pos;
      INT_2U					m_compression_scheme;
      INT_2U					m_compression_level;

      template< typename T >
      const FrameSpec::Object*
      object_ptr( General::SharedPtr< T > Obj );

      template< typename T >
      const FrameSpec::Object*
      object_ptr( const T* Obj );
    };

    //-------------------------------------------------------------------
    //-------------------------------------------------------------------
    
    inline OStream::Pointer::
    Pointer( object_type Data )
      : m_pointer( Data )
    {
    }

    inline OStream::object_type OStream::Pointer::
    GetPointer(  ) const
    {
      return m_pointer;
    }

    //-------------------------------------------------------------------
    //-------------------------------------------------------------------

    inline OStream::
    OStream( buffer_type* Buffer )
      : std::ostream( dynamic_cast< std::streambuf* >( Buffer ) ),
	m_buffer( Buffer ),
	m_compression_scheme( 0 ),
	m_compression_level( 0 )
    {
      if ( Buffer )
      {
	m_filter_internally = Buffer->FilterInternally( );
      }
      m_write_objects_pos = m_write_objects.begin( );
    }

    inline OStream::
    ~OStream( )
    {
    }

    inline void OStream::
    Compression( INT_2U& CompressionScheme, INT_2U& CompressionLevel )
    {
      const INT_2U old_scheme( m_compression_scheme );
      const INT_2U old_level( m_compression_level );

      m_compression_scheme = CompressionScheme;
      m_compression_level = CompressionLevel;

      CompressionScheme = old_scheme;
      CompressionLevel = old_level;
    }

    inline INT_2U OStream::
    CompressionScheme( ) const
    {
      return m_compression_scheme;
    }

    inline INT_2U OStream::
    CompressionLevel( ) const
    {
      return m_compression_level;
    }

    inline CheckSumFilter* OStream::
    GetCheckSumFile( ) const
    {
      return m_file_checksum.get( );
    }

    inline void OStream::
    SetCheckSumFile( const CheckSum::kind_type Type )
    {
      if ( m_file_checksum.get( ) )
      {
	rdbuf( )->FilterRemove( m_file_checksum.get( ) );
      }
      if ( Type == CheckSum::NONE )
      {
	m_file_checksum.reset( (CheckSumFilter*)NULL );
      }
      else
      {
	m_file_checksum.reset( new CheckSumFilter( Type ) );
	rdbuf( )->FilterAdd( m_file_checksum.get( ) );
      }
    }

    inline CheckSumFilter* OStream::
    GetCheckSumFrame( ) const
    {
      return m_frame_checksum.get( );
    }

    inline void OStream::
    SetCheckSumFrame( const CheckSum::kind_type Type )
    {
      if ( m_frame_checksum.get( ) )
      {
	rdbuf( )->FilterRemove( m_frame_checksum.get( ) );
      }
      if ( Type == CheckSum::NONE )
      {
	m_frame_checksum.reset( (CheckSumFilter*)NULL );
      }
      else
      {
	m_frame_checksum.reset( new CheckSumFilter( Type ) );
	rdbuf( )->FilterAdd( m_frame_checksum.get( ) );
      }
    }

    inline CheckSumFilter* OStream::
    GetCheckSumObject( ) const
    {
      return m_obj_checksum.get( );
    }

    inline void OStream::
    SetCheckSumObject( const CheckSum::kind_type Type )
    {
      if ( m_obj_checksum.get( ) )
      {
	rdbuf( )->FilterRemove( m_obj_checksum.get( ) );
      }
      if ( Type == CheckSum::NONE )
      {
	m_obj_checksum.reset( (CheckSumFilter*)NULL );
      }
      else
      {
	m_obj_checksum.reset( new CheckSumFilter( Type ) );
	rdbuf( )->FilterAdd( m_obj_checksum.get( ) );
      }
    }

    template < typename Header >
    inline const Header* OStream::
    GetFrHeader( ) const
    {
      return dynamic_cast< const Header * >( m_file_header.get( ) );
    }

    inline OStream& OStream::
    write( const char_type* S, std::streamsize N )
    {
      if ( ! m_filter_internally )
      {
	m_buffer->filter( S, S + N );
      }
      ostream_type::write( S, N );
      return *this;
    }

    //-------------------------------------------------------------------
    /// This method retieves the next object to be written along with
    ///  information indicating if the object should be deallocated
    ///  once written.
    ///
    /// \param[out] Obj
    ///     This is a pointer to the object which should be written to
    ///	    the output stream.
    /// \param[out] Delete
    ///     Boolean indicating if Obj should be deleted after writing.
    ///
    /// \return
    ///     If true, then Obj and Delete were set; false otherwise
    //-------------------------------------------------------------------
    inline OStream::object_type OStream::
    Pop( )
    {
      object_type	retval;

      //-----------------------------------------------------------------
      // Check if there is anything to be done
      //-----------------------------------------------------------------

      if ( m_write_objects.empty( ) == false )
      {
	//---------------------------------------------------------------
	// Always take from the front of the list
	//---------------------------------------------------------------

	retval = m_write_objects.front( );

	//---------------------------------------------------------------
	// Remove the information from the list and reset the insertion
	//  iterator.
	//---------------------------------------------------------------

	m_write_objects_pos
	  = m_write_objects.erase( m_write_objects.begin( ) );

      }

      return retval;
    }

    //-------------------------------------------------------------------
    /// This method adds an object to the queue of objects to be written.
    /// It preserves the order in which the objects should be written
    ///  to the stream by keeping a local iterator indicating the point
    ///  of insertion.
    ///
    /// \param[in] Obj
    ///     The object to be written.
    /// \param[in] Delete
    ///     A boolean indicating if the object should be deleted once
    ///      written to the stream.
    //-------------------------------------------------------------------
    inline void OStream::
    PushSingle( object_type Obj )
    {
      if ( ( Obj )
	   && ( ! m_dictionary.Ref( Obj ) ) )
      {
	m_dictionary.RefCreate( Obj, *this );
	// Ensure a valid value when starting
	if ( m_write_objects.size( ) == 0 )
	{
	  m_write_objects_pos = m_write_objects.begin( );
	}
	// Insert the information into the queue
	m_write_objects_pos =
	  m_write_objects.insert( m_write_objects_pos, Obj );
	// Advance the iterator to just past the object inserted.
	++m_write_objects_pos;
      }
    }

    inline OStream& OStream::
    operator<< ( object_type Data )
    {
      PushSingle( Data );
      const PTR_STRUCT_BASE& ptr( Reference( Data ) );
      ptr.Write( *this );
      return *this;
    }

    inline OStream& OStream::
    operator<< ( Pointer Data )
    {
      PushSingle( Data.GetPointer( ) );
      const PTR_STRUCT_BASE& ptr( Reference( Data.GetPointer( )) );
      ptr.Write( *this );
      return *this;
    }

    template < >
    inline OStream& OStream::
    operator<<( const General::GPSTime& Data )
    {
      *this << Data.GetSeconds( )
	    << Data.GetNanoseconds( );
      return *this;
    }

    template < typename DT >
    inline OStream& OStream::
    operator<<( const std::vector< DT >& Data )
    {
      if ( Data.size( ) > 0 )
      {
	return write( reinterpret_cast< const char_type* >( &(Data[0]) ),
		      sizeof( DT ) * Data.size( ) );
      }
      return *this;
    }

    template < typename DT >
    inline OStream& OStream::
    operator<<( const DT& Data )
    {
      return write( reinterpret_cast< const char_type* >( &Data ),
		    sizeof( Data ) );
    }

    inline INT_4U OStream::
    GetNumberOfFrames( ) const
    {
      return 0;
    }

    inline void OStream::
    setFrHeader( std::auto_ptr< FrameSpec::Object >& Obj )
    {
      m_file_header = Obj;
    }

    template< typename T >
    inline const FrameSpec::Object* OStream::
    object_ptr( typename General::SharedPtr< T > Obj )
    {
      return Obj.get( );
    }

    template< >
    inline const FrameSpec::Object* OStream::
    object_ptr< FrStatData >(  const FrStatData* Obj )
    {
      return reinterpret_cast< const FrameSpec::Object* >( Obj );
    }
  }
}

#endif /* FrameCPP__COMMON__IOStream_HH */
